
from . import aixm
from . import gml
from . import geocalc


class AixmGeometryResolver:
    """
    This class converts AIXM gemoetries (Arc, Circle, ..) into
    single points. This means, after the conversion, the
    airspaces will only contains points::

        parser = AixmAirspaceParser.AixmAirspaceParser()
        airspaces, borders = parser.parse("airspace-aixm.xml")
        resolver = AixmGeometryResolver()
        resolver.resolve_airspaces(airspaces)
        converter = AixmAirspaceParser.AixmOpenairConverter()
        openairs = converter.convert_airspaces(airspaces)

        fp = open("airspace-openair.txt", 'wb')
        writer = aerofiles.openair.writer.Writer(fp)
        for openair in openairs:
            writer.write_record(openair)

    `arc_angle_step` is the number of degrees, which will be used to convert
    arcs into points. A value of `1` means, that a circle has 360 points.
    """

    def __init__(self, arc_angle_step=1):
        self.arc_angle_step = arc_angle_step

    def compute_Arc(self, center: gml.Position, radius_km, start_angle, end_angle, clockwise, use_edge):
        elements = []

        if clockwise:
            reverse = False
        else:
            reverse = True
            clockwise = True
            tmp = start_angle
            start_angle = end_angle
            end_angle = tmp

        dir = self.arc_angle_step

        if clockwise:
            while start_angle > end_angle:
                end_angle = end_angle + 360
        else:
            dir = -dir
            while start_angle < end_angle:
                start_angle = start_angle + 360

        # icstart_angle)
        # ic(end_angle)
        # ic(clockwise)

        if not use_edge:
            start_angle = start_angle + dir
            end_angle = end_angle - dir

        angle = start_angle

        while True:
            # print("loop angle:", angle)
            if clockwise:
                if angle >= end_angle:
                    angle = end_angle
                    break
            else:
                if angle <= end_angle:
                    angle = end_angle
                    break

            (lat, lon) = geocalc.geo_destination(center, angle, radius_km)
            # print(f'lat={lat} lon={lon}')
            element = gml.Position(latitude=lat, longitude=lon)
            elements.append(element)
            angle = angle + dir

        if reverse:
            elements.reverse()

        return elements

    def resolve_DB(self, center: gml.Position, start: gml.Position, end: gml.Position, clockwise):

        (dist_s, bearing_s) = geocalc.geo_distance(center, start)
        (dist_e, bearing_e) = geocalc.geo_distance(center, end)

        dist_s_km = (dist_s / 100) / 1000

        elements = self.compute_Arc(
            center, dist_s_km, bearing_s, bearing_e, clockwise, False)
        elements.insert(0, start)
        elements.append(end)

        return elements

    def resolve_PointSegment(self, segment):
        # ic(segment)
        positions = []
        positions.append(segment.point)
        return gml.ResolvedSegment(positions=positions, parent=segment)

    def resolve_CircleSegment(self, segment):
        # ic(segment)
        positions = []
        radius_km = float(geocalc.nautical_miles_to_km(segment.radius))
        positions = self.compute_Arc(
            segment.center, radius_km, 0, 180, True, True)
        positions.extend(self.compute_Arc(
            segment.center, radius_km, 180, 0, True, True))
        return gml.ResolvedSegment(positions=positions, parent=segment)

    def resolve_ArcSegment(self, segment):
        # ic(segment)
        positions = []
        radius_km = float(geocalc.nautical_miles_to_km(segment.radius))
        positions = self.compute_Arc(
            segment.center, radius_km, segment.start_bearing, segment.end_bearing, segment.clockwise, True)
        return gml.ResolvedSegment(positions=positions, parent=segment)

    def resolve_ResolvedSegment(self, segment):
        return segment

    def resolve_segment(self, segment: gml.GeometrySegment):
        type_name = segment.__class__.__name__
        method_name = f'resolve_{type_name}'
        method = getattr(self, method_name)
        return method(segment)

    def resolve_curve(self, curve: gml.Curve):
        if curve.__class__.__name__ == "CurveRef":
            print(f'Skipping {curve}')
            return

        segments_resolved = []
        for segment in curve.segments:
            segments_resolved.append(self.resolve_segment(segment))
        curve.segments = segments_resolved

    def resolve_volume(self, volume: aixm.AirspaceVolume):
        for curve in volume.curves:
            self.resolve_curve(curve)

    def resolve_airspace(self, airspace):
        for component in airspace.components:
            self.resolve_volume(component.volume)

    def resolve_airspaces(self, airspaces):
        for airspace in airspaces:
            self.resolve_airspace(airspace)

    def find_border(self, borders, gml_id):
        for border in borders:
            # ic(gml_id, border.gml_id)
            if border.gml_id == gml_id:
                return border
        return None



#!/usr/bin/python
#
# Convert AIXM airspaces into openair.
#
# Walk though all aixm.Airspace and convert them to openair.
#
# This script is licensed under GPLv2.
#
# 2025-07-02, tilmann@bubecks.de
#

import sys

from . import aixm
from . import gml
from . import geocalc

# from icecream import ic


class AixmOpenairConverter:
    """
    Convert all AIXM airspaces returned from AixmAirspaceParser into openair records.
    These can then be written by `aerofiles.openair.writer`::

        parser = AixmAirspaceParser.AixmAirspaceParser()
        airspaces, borders = parser.parse("airspace-aixm.xml")
        converter = AixmAirspaceParser.AixmOpenairConverter()
        openairs = converter.convert_airspaces(airspaces)

        fp = open("airspace-openair.txt", 'wb')
        writer = aerofiles.openair.writer.Writer(fp)
        for openair in openairs:
            writer.write_record(openair)

    """

    def __init__(self, airspaces, borders, convert_DA_to_DB=False):
        self.airspaces = airspaces
        self.borders = borders
        self.convert_DA_to_DB = convert_DA_to_DB

    def convert_PointSegment(self, segment: gml.PointSegment):
        # ic(segment)
        element = dict()
        element["type"] = "point"
        element["location"] = [segment.point.latitude, segment.point.longitude]

        return [element]

    def convert_CircleSegment(self, segment):
        # ic(segment)
        element = dict()
        element["type"] = "circle"
        element["center"] = [segment.center.latitude, segment.center.longitude]
        element["radius"] = segment.radius

        return [element]

    def convert_ArcSegment(self, segment):
        # ic(segment)
        element = dict()
        element["type"] = "arc"
        element["center"] = [segment.center.latitude, segment.center.longitude]
        element["clockwise"] = segment.clockwise
        if self.convert_DA_to_DB:
            element["start"] = geocalc.geo_destination(segment.center,
                                                       segment.start_bearing,
                                                       float(geocalc.nautical_miles_to_km(segment.radius)))
            element["end"] = geocalc.geo_destination(segment.center,
                                                     segment.end_bearing,
                                                     float(geocalc.nautical_miles_to_km(segment.radius)))
        else:
            element["start"] = segment.start_bearing
            element["end"] = segment.end_bearing
            element["radius"] = segment.radius

        return [element]

    def convert_ResolvedSegment(self, segment):
        elements = []
        for position in segment.positions:
            element = dict()
            element["type"] = "point"
            element["location"] = [position.latitude, position.longitude]
            elements.append(element)
        return elements

    def convert_segment(self, segment):
        type_name = segment.__class__.__name__
        method_name = f'convert_{type_name}'
        method = getattr(self, method_name)
        return method(segment)

    def find_airspace(self, identifier):
        for airspace in self.airspaces:
            if airspace.gml_id == identifier:
                return airspace
        return None

    def convert_volume(self, volume: aixm.AirspaceVolume):
        elements = []
        if len(volume.dependencies) > 0:
            assert (len(volume.dependencies) == 1)
            dependency = volume.dependencies[0]
            airspace = self.find_airspace(dependency.xlink)
            if airspace is not None:
                print(f'airspace {dependency.xlink} not found')
                sys.exit(1)
            print(len(airspace.components))
            assert (len(airspace.components) == 1)
            volume = airspace.components[0].volume
            # ic(volume)
            # sys.exit(0)

        for curve in volume.curves:
            if curve.__class__.__name__ == "CurveRef":
                print(f'Skipping {curve} [not implemented]')
            else:
                for segment in curve.segments:
                    elements.extend(self.convert_segment(segment))
        return elements

    def convert_vertical_limit(self, limit, limit_reference):
        if limit.uom == "FL":
            if limit_reference != "STD":
                print("Unknown limit reference", limit.dump())
                sys.exit(1)
            return f'FL{limit.value}'

        if limit_reference == "MSL":
            limit_reference = "AMSL"
        if limit_reference == "SFC":
            limit_reference = "AGL"

        if limit.value == "0" and limit_reference == "AGL":
            return "GND"

        return f'{limit.value} {limit.uom} {limit_reference}'

    def get_dependant_components(self, airspace):
        components = []
        for component in airspace.components:
            volume = component.volume
            if len(volume.dependencies) > 0:
                assert (len(volume.curves) == 0)
                for volume_dependency in volume.dependencies:
                    airspace_dep = self.find_airspace(volume_dependency.xlink)
                    if airspace_dep is None:
                        print(
                            f'Unable to find dependant airspace "{volume_dependency.xlink}" for AirspaceVolume "{volume.gml_id}"')
                        continue
                    airspace_dep.is_referenced = True
                    components.extend(
                        self.get_dependant_components(airspace_dep))
            else:
                components.append(component)

        return components

    def convert_border(self, border):

        openairs = []

        # ic(border)
        openair = dict()
        openair["type"] = "airspace"
        openair["class"] = "BORDER"
        openair["name"] = border.name
        openair["ident"] = border.gml_id
        openair["ceiling"] = "GND"
        openair["floor"] = "1000 FT AGL"

        elements = []
        for segment in border.curve.segments:
            elements.extend(self.convert_segment(segment))
        openair["elements"] = elements

        openairs.append(openair)

        return openairs

    def convert_airspace(self, airspace):

        openairs = []

        # ic(airspace)
        openair = dict()
        openair["type"] = "airspace"
        openair["class"] = airspace.class_airspace
        openair["airspace_type"] = airspace.type_airspace
        openair["ground_name"] = airspace.designator
        openair["name"] = airspace.name
        openair["ident"] = airspace.gml_id
        openair["elements"] = []
        # ic(openair["class"])

        components = self.get_dependant_components(airspace)

        if len(components) < 1:
            print(f'No components in {airspace}')
            sys.exit(1)

        volume_base = components[0].volume

        # ic(airspace)
        openair["ceiling"] = self.convert_vertical_limit(
            volume_base.upper_limit, volume_base.upper_limit_reference)
        openair["floor"] = self.convert_vertical_limit(
            volume_base.lower_limit, volume_base.lower_limit_reference)

        for component in components:
            oa = openair.copy()
            oa["elements"] = self.convert_volume(component.volume)
            openairs.append(oa)

        # ic(openairs)
        return openairs

    def convert_borders(self):
        openairs = []

        for border in self.borders:
            openairs.extend(self.convert_border(border))

        return openairs

    def convert_airspaces(self, only_referenced_airspaces=True):
        openairs = []

        for airspace in self.airspaces:
            if not only_referenced_airspaces or not airspace.is_referenced:
                oa = self.convert_airspace(airspace)
                if oa is not None:
                    openairs.extend(oa)

        return openairs



#!/usr/bin/python
#
# 2025-07-01, tilmann@bubecks.de
#

import decimal
import sys
import xml.etree.ElementTree as ET

# from icecream import ic

from . import aixm
from . import gml

writer = None

GML = "{http://www.opengis.net/gml/3.2}"
AIXM = "{http://www.aixm.aero/schema/5.1.1}"
XLINK = "{http://www.w3.org/1999/xlink}"


class AixmAirspaceParser:
    """
    Parse an AIXM 5.1.1 document to find all Airspaces::

        parser = AixmAirspaceParser.AixmAirspaceParser()
        airspaces, borders = parser.parse(args.xml)
        parser.resolve()
        if parser.contains_curveref:
            print(f'WARN: the AIXM file contains references curves. For "Germany" we recommend "ED_Airspace_StrokedBorders...xml"')

    Use `parser.resolve()` to resolve all referenced Airspaces.
    """

    def __init__(self):
        self.contains_curveref = False
        self.airspaces = []
        self.borders = []

    def parse_pos(self, pos):
        p = pos.split()
        # ic(p)
        return [float(p[0]), float(p[1])]
        # return [decimal.Decimal(p[0]), decimal.Decimal(p[1])]

    def parseGeodesigString(self, gds):
        positions = []

        for element in gds.findall("./"):
            if element.tag == GML + "pos":
                # ic(p)
                ll = self.parse_pos(element.text)
                p = gml.PointSegment(gml.Position(ll[0], ll[1]))
                positions.append(p)
            elif element.tag == GML + "posList":
                # ic(p)
                p = element.text.split()
                if len(p) % 2 != 0:
                    print("ERROR: keine paare", element.text)
                    sys.exit(1)
                for i in range(int(len(p) / 2)):
                    ps = gml.PointSegment(gml.Position(
                        float(p[i * 2]), float(p[i * 2 + 1])))
                    positions.append(ps)
            else:
                print("Unbekanntest GeodesigElement", element.tag)
                sys.exit(1)

        return positions

    def parseCircleByCenterPoint(self, circle2):
        pos = circle2.find(GML + "pos")
        radius = circle2.find(GML + "radius")

        ll = self.parse_pos(pos.text)
        circle = gml.CircleSegment(center=gml.Position(
            ll[0], ll[1]), radius=decimal.Decimal(radius.text))

        return circle

    def parseArcByCenterPoint(self, circle):
        pos = circle.find(GML + "pos")
        radius = circle.find(GML + "radius")

        ll = self.parse_pos(pos.text)

        start_bearing = decimal.Decimal(circle.find(GML + "startAngle").text)
        end_bearing = decimal.Decimal(circle.find(GML + "endAngle").text)
        arc = gml.ArcSegment(center=gml.Position(ll[0], ll[1]),
                             radius=decimal.Decimal(radius.text),
                             start_bearing=start_bearing,
                             end_bearing=end_bearing,
                             clockwise=start_bearing < end_bearing
                             )

        return arc

    def parseSegments(self, segments_element):
        segments = []
        for element in segments_element.findall("./"):
            # ic(element.tag)
            if "GeodesicString" in element.tag:
                segments.extend(self.parseGeodesigString(element))
            elif "LineStringSegment" in element.tag:
                segments.extend(self.parseGeodesigString(element))
            elif "CircleByCenterPoint" in element.tag:
                segments.append(self.parseCircleByCenterPoint(element))
            elif "ArcByCenterPoint" in element.tag:
                segments.append(self.parseArcByCenterPoint(element))
            else:
                print("Unbekanntes Segment", element.tag)
                sys.exit(1)

        return segments

    def parseCurve(self, curve_element):
        curve = gml.CurveImpl()
        for segment in curve_element.iter(GML + "segments"):
            segments = self.parseSegments(segment)
            curve.add_segments(segments)
        return curve

    def parseCurveMember(self, curveMember):
        curves = []
        href = curveMember.get(XLINK + "href")
        if href is not None:
            self.contains_curveref = True
            # ET.dump(curveMember)
            if href.startswith("urn:uuid:"):
                href = href[9:]
            curve = gml.CurveRef(href)
            return curve

        count = 0
        for curve in curveMember.iter(GML + "Curve"):
            count += 1
            c = self.parseCurve(curve)
            curves.append(c)

        assert (count == 1)

        return curves[0]

    def parseRing(self, ring):
        curves = []
        for member in ring.iter(GML + "curveMember"):
            c2 = self.parseCurveMember(member)
            curves.append(c2)

        return curves

    def parseAirspaceVolume(self, volume_element):

        volume = aixm.AirspaceVolume()
        # ET.dump(volume_element)
        volume.gml_id = volume_element.get(GML + "id")
        # sys.exit(0)

        contributorAirspace = volume_element.find(
            ".//" + AIXM + "AirspaceVolumeDependency")
        if contributorAirspace is not None:
            # ic("CONTRIBUTOR")
            # ET.dump(contributorAirspace)
            for airspace in contributorAirspace.iter(AIXM + "theAirspace"):
                # ET.dump(airspace)
                # ic(airspace.keys())
                dependency = volume_element.find(".//" + AIXM + "dependency")
                href = airspace.get(XLINK + "href")
                if href.startswith("urn:uuid:"):
                    href = href[9:]
                avd = aixm.AirspaceVolumeDependency()
                avd.type_dependency = dependency.text
                avd.xlink = href
                volume.add_dependency(avd)

                # ic(avd)
                # sys.exit(0)
                return volume

        (volume.upper_limit, volume.upper_limit_reference) = self.parseLimit(
            volume_element, "upper")
        (volume.lower_limit, volume.lower_limit_reference) = self.parseLimit(
            volume_element, "lower")

        # Ensure, that we have the following elements. Our logic depends on this.
        # If not, then Exception will be thrown
        ring = (
            volume_element.find(AIXM + "horizontalProjection")
            .find(AIXM + "Surface")
            .find(GML + "patches")
            .find(GML + "PolygonPatch")
            .find(GML + "exterior")
            .find(GML + "Ring")
        )

        count = 0
        for ring in volume_element.iter(GML + "Ring"):
            count += 1
            curves = self.parseRing(ring)
            # elements.extend(segments)
            volume.add_curves(curves)
        assert (count == 1)

        return volume

    def parseLimit(self, airspace, low_or_up):
        key = low_or_up + "Limit"
        limit_element = airspace.find(AIXM + key)
        if limit_element is not None:
            uom = limit_element.get("uom")
            limitRef = airspace.find(AIXM + key + "Reference")
            ref = limitRef.text
            value = limit_element.text
            return (aixm.ValDistanceVertical(value=value, uom=uom), ref)

        return (None, None)

    def parseTimesheet(self, timesheet_element):
        time_reference = timesheet_element.find(AIXM + "timeReference")
        start_date = timesheet_element.find(AIXM + "startDate")
        end_date = timesheet_element.find(AIXM + "endDate")
        start_time = timesheet_element.find(AIXM + "startTime")
        end_time = timesheet_element.find(AIXM + "endTime")
        day = timesheet_element.find(AIXM + "day")
        daylightSavingAdjust = timesheet_element.find(
            AIXM + "daylightSavingAdjust").text
        daylightSavingAdjust = daylightSavingAdjust == "YES"

        timesheet = aixm.AixmTimesheet(time_reference=time_reference.text,
                                       start_date=start_date.text,
                                       end_date=end_date.text,
                                       start_time=start_time.text,
                                       end_time=end_time.text,
                                       daylight_saving_adjust=daylightSavingAdjust,
                                       day=day.text
                                       )
        return timesheet

    def parseAirspaceActivation(self, element):
        activation = aixm.AirspaceActivation()
        activation.activity = element.find(AIXM + "activity").text
        activation.status = element.find(AIXM + "status").text

        count = 0
        for timesheet in element.iter(AIXM + "Timesheet"):
            count += 1
            sheet = self.parseTimesheet(timesheet)
            if sheet is not None:
                # ic(sheet)
                activation.add_timesheet(sheet)

        # AirspaceActivation without Timesheet is invalid
        if count == 0:
            return None

        return activation

    def parseAirspaceGeometryComponent(self, airspace: aixm.Airspace, airspaceGeometryComponent):

        component = aixm.AirspaceGeometryComponent()
        operation = airspaceGeometryComponent.find(AIXM + "operation")
        operationSequence = airspaceGeometryComponent.find(
            AIXM + "operationSequence")
        if operation is not None and operationSequence is not None:
            component.set_operation(
                operation.text, int(operationSequence.text))

        count = 0
        for volume in airspaceGeometryComponent.iter(AIXM + "AirspaceVolume"):
            count += 1
            # ic(count)
            component.set_volume(self.parseAirspaceVolume(volume))
            # if len(positions) == 0:
            #    ic("Empty positions", openair)
            #    sys.exit(1)
            #    return
            # openair["elements"] = positions
            # ic(positions)

        assert (count == 1)

        return component

    def parseAirspace(self, element):

        airspace = aixm.Airspace()

        airspace.gml_id = element.find(GML + "identifier").text
        airspace.designator = element.find(".//" + AIXM + "designator").text
        airspace.name = element.find(".//" + AIXM + "name").text
        airspace.type_airspace = element.find(".//" + AIXM + "localType").text
        airspace.class_airspace = element.find(".//" + AIXM + "type").text
        if airspace.class_airspace == "CLASS":
            classification = element.find(".//" + AIXM + "classification")
            if classification is not None:
                airspace.class_airspace = classification.text

        # ic(as_name, as_type)
        # ic(airspace)

        count = 0
        for airspaceGeometryComponent in element.iter(AIXM + "AirspaceGeometryComponent"):
            count += 1
            airspace.add_component(self.parseAirspaceGeometryComponent(
                airspace, airspaceGeometryComponent))

        # ic(airspace)

        count = 0
        for activation in element.iter(AIXM + "AirspaceActivation"):
            act = self.parseAirspaceActivation(activation)
            if act is not None:
                # ic(sheet)
                count += 1
                airspace.activation = act

        assert count <= 1, f'More than 1 AirspaceActivation in {airspace}'

        return airspace

    def parseGeoBorder(self, geoBorder):
        gml_id = geoBorder.find(GML + "identifier").text
        geo_name = geoBorder.find(".//" + AIXM + "name").text
        curve = geoBorder.find(".//" + AIXM + "Curve")

        border = aixm.GeoBorder(gml_id, geo_name)
        border.curve = self.parseCurve(curve)

        return border

    def find_airspace(self, identifier):
        for airspace in self.airspaces:
            if airspace.gml_id == identifier:
                return airspace
        return None

    def get_dependant_components(self, airspace):
        components = []
        for component in airspace.components:
            volume = component.volume
            if len(volume.dependencies) > 0:
                assert (len(volume.curves) == 0)
                for volume_dependency in volume.dependencies:
                    airspace_dep = self.find_airspace(volume_dependency.xlink)
                    if airspace_dep is None:
                        print(
                            f'Unable to find dependant airspace "{volume_dependency.xlink}" for AirspaceVolume "{volume.gml_id}"')
                        continue
                    airspace_dep.is_referenced = True
                    components.extend(
                        self.get_dependant_components(airspace_dep))
            else:
                components.append(component)

        return components

    def mark_referenced_airspaces(self):
        for airspace in self.airspaces:
            _ = self.get_dependant_components(airspace)

    def resolve_airspace(self, airspace):
        # ic(airspace)
        airspace.components = self.get_dependant_components(airspace)

    def resolve(self):
        """
         Walk through all airspaces and resolve airspace.component.  If
         an airspace has AirspaceVolumes with dependencies, these will
         be collected and replaced with the referenced Volumes.

         This means, that after calling this method, all airspaces
         will have only airspace.components and no dependencies. This
         will ease the usage of airspaces by following methods.
        """

        for airspace in self.airspaces:
            self.resolve_airspace(airspace)

    def parse(self, xml):
        """

        Parse the given AXML 5.1.1 document and returns all airspaces and borders.
        """

        self.borders = []
        self.airspaces = []

        tree = ET.parse(xml)
        root = tree.getroot()

        for element in root.iter(AIXM + "GeoBorder"):
            border = self.parseGeoBorder(element)
            # ic(border)
            self.borders.append(border)

        for airspace in root.iter(AIXM + "Airspace"):
            airspace = self.parseAirspace(airspace)
            if airspace is not None:
                self.airspaces.append(airspace)

        self.mark_referenced_airspaces()

        return (self.airspaces, self.borders)



# flake8: noqa

# from .reader import Reader, LowLevelReader
# from .writer import Writer
from .AixmAirspaceParser import AixmAirspaceParser
from .AixmOpenairConverter import AixmOpenairConverter
from .AixmGeometryResolver import AixmGeometryResolver

# from .aixm import AixmTimesheet



from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class Position:
    """Geographic position in decimal degrees"""
    latitude: float
    longitude: float

    def to_dms(self, coord_type: str) -> str:
        """Convert to degrees/minutes/seconds format for OpenAir"""
        value = self.latitude if coord_type == 'lat' else self.longitude
        abs_value = abs(value)
        degrees = int(abs_value)
        minutes = int((abs_value - degrees) * 60)
        seconds = ((abs_value - degrees) * 60 - minutes) * 60

        if coord_type == 'lat':
            direction = 'N' if value >= 0 else 'S'
        else:
            direction = 'E' if value >= 0 else 'W'

        return f"{degrees:02d}:{minutes:02d}:{seconds:05.2f} {direction}"


@dataclass
class GeometrySegment:
    """Base class for geometry segments"""
    pass


@dataclass
class PointSegment(GeometrySegment):
    """Circular airspace boundary"""
    point: Position


@dataclass
class CircleSegment(GeometrySegment):
    """Circular airspace boundary"""
    center: Position
    radius: float  # in nautical miles


@dataclass
class ArcSegment(GeometrySegment):
    """Arc segment of airspace boundary"""
    center: Position
    radius: float
    start_bearing: float
    end_bearing: float
    clockwise: bool = True


@dataclass
class LineSegment(GeometrySegment):
    """Line segment between two points"""
    start: Position
    end: Position


@dataclass
class ResolvedSegment(GeometrySegment):
    positions: List['Position'] = field(default_factory=list)
    parent: Optional[GeometrySegment] = None


@dataclass
class Curve:
    """Base class for geometry segments"""
    pass


@dataclass
class CurveRef(Curve):
    ref_id: str


@dataclass
class CurveImpl(Curve):
    segments: List[GeometrySegment] = field(default_factory=list)

    def add_segment(self, segment):
        self.segments.append(segment)

    def add_segments(self, segments):
        self.segments.extend(segments)


@dataclass
class Ring:
    curves: List[Curve] = field(default_factory=list)



import decimal
import math

# Converter function to convert nautical miles to km
#
# @param nm nautical miles
#
# @return corresponding value in km


def nautical_miles_to_km(nm):
    return nm * decimal.Decimal(1.852)


def geo_destination(p, angle, distance_km: float):
    lat1 = p.latitude
    lon1 = p.longitude

    angle = math.radians(angle)
    dx = math.sin(angle) * distance_km
    dy = math.cos(angle) * distance_km

    (kx, ky) = get_kx_ky(lat1)

    lon2 = lon1 + dx / kx
    lat2 = lat1 + dy / ky

    return (lat2, lon2)


# Compute the distance between two GPS points in 2 dimensions
# (without altitude). Latitude and longitude parameters must be given
# as fixed integers multiplied with GPS_COORD_MULT.
#
# \param lat1 the latitude of the 1st GPS point
# \param lon1 the longitude of the 1st GPS point
# \param lat2 the latitude of the 2nd GPS point
# \param lon2 the longitude of the 2nd GPS point
# \param FAI use FAI sphere instead of WGS ellipsoid
# \param bearing pointer to bearing (NULL if not used)
#
# \return the distance in cm.
#
def geo_distance(p1, p2):

    lat1 = p1.latitude
    lon1 = p1.longitude
    lat2 = p2.latitude
    lon2 = p2.longitude

    d_lon = (lon2 - lon1)
    d_lat = (lat2 - lat1)

    # DEBUG("#d_lon=%0.10f\n", d_lon);
    # DEBUG("#d_lat=%0.10f\n", d_lat);

    # DEBUG("lat1=%li\n", lat1);
    # DEBUG("lon1=%li\n", lon1);
    # DEBUG("lat2=%li\n", lat2);
    # DEBUG("lon2=%li\n", lon2);

    # WGS
    # DEBUG("f=0\n");
    # DEBUG("#lat=%0.10f\n", (lat1 + lat2) / ((float)GPS_COORD_MUL * 2));

    (kx, ky) = get_kx_ky((lat1 + lat2) / 2)

    # DEBUG("#kx=%0.10f\n", kx);
    # DEBUG("#ky=%0.10f\n", ky);

    d_lon = d_lon * kx
    d_lat = d_lat * ky

    # DEBUG("#d_lon=%0.10f\n", d_lon);
    # DEBUG("#d_lat=%0.10f\n", d_lat);

    dist = math.sqrt(math.pow(d_lon, 2) + math.pow(d_lat, 2)) * 100000.0

    if d_lon == 0 and d_lat == 0:
        bearing = 0
    else:
        bearing = (math.degrees(math.atan2(d_lon, d_lat)) + 360) % 360
        # DEBUG("a=%d\n", *bearing);

    # DEBUG("d=%lu\n\n", dist);

    return (dist, bearing)


def get_kx_ky(lat):
    fcos = math.cos(math.radians(lat))
    cos2 = 2. * fcos * fcos - 1.
    cos3 = 2. * fcos * cos2 - fcos
    cos4 = 2. * fcos * cos3 - cos2
    cos5 = 2. * fcos * cos4 - cos3
    # multipliers for converting longitude and latitude
    # degrees into distance (http://1.usa.gov/1Wb1bv7)
    kx = (111.41513 * fcos - 0.09455 * cos3 + 0.00012 * cos5)
    ky = (111.13209 - 0.56605 * cos2 + 0.0012 * cos4)
    return (kx, ky)


def decimal_degrees_to_dms(decimal_degrees):
    # mnt,sec = divmod(decimal_degrees*3600,60)
    # deg,mnt = divmod(mnt, 60)
    # return (round(deg),round(mnt),round(sec))

    # decimals, number = math.modf(decimal_degrees)
    # deg = int(number)
    # mnt = round(decimals * 60)
    # sec = (decimal_degrees - deg - mnt / 60) * 3600.00
    # return deg,mnt,round(sec)

    decimals, number = math.modf(decimal_degrees)
    deg = int(number)
    mnt = int(decimals * 60)
    sec = round((decimal_degrees - deg - mnt / 60) * 3600.00)
    if sec == 60:
        sec = 0
        mnt += 1
        if mnt == 60:
            mnt = 0
            deg += 1
    return deg, mnt, sec


def strDegree(v, width):
    (degrees, minutes, seconds) = decimal_degrees_to_dms(abs(v))
    return f'{degrees:0{width}d}:{minutes:02d}:{seconds:02d}'


def strLatLon(P):
    result = strDegree(abs(P[0]), 2) + " "
    if P[0] >= 0:
        result = result + "N "
    else:
        result = result + "S "

    result = result + strDegree(abs(P[1]), 3) + " "
    if P[1] >= 0:
        result = result + "E "
    else:
        result = result + "W "
    return result



#!/usr/bin/python
#
# Convert AIXM 5.1.1 XML files into openair airspaces.
#
# This script is licensed under GPLv2.
#
# 2025-07-02, tilmann@bubecks.de
#

import argparse

import aerofiles.openair.writer
from aerofiles.aixm import AixmAirspaceParser, AixmGeometryResolver, AixmOpenairConverter


def main():
    parser = argparse.ArgumentParser(
        description='Convert AIXM 5.1.1 airspaces into openair')
    parser.add_argument('-x', '--xml', required=True, help='AIXM filename to read from',
                        default="ED_Airspace_StrokedBorders_2024-03-21_2024-03-21_snapshot.xml")
    parser.add_argument('-o', '--output', required=True,
                        help='openair filename for ourpur')
    parser.add_argument('-r', '--resolve', required=False,
                        help='Resolve all arcs to points with given degree step.')
    parser.add_argument('-b', '--borders', required=False,
                        action='store_true', help='Include GeoBorder into openair output')
    args = parser.parse_args()

    parser = AixmAirspaceParser()

    airspaces, borders = parser.parse(args.xml)

    if parser.contains_curveref:
        print('WARN: the AIXM file contains references curves. For "Germany" we recommend "ED_Airspace_StrokedBorders...xml"')

    if args.resolve:
        resolver = AixmGeometryResolver(int(args.resolve))
        resolver.resolve_airspaces(airspaces)
        # resolver.crop_borders(airspaces, borders)

    converter = AixmOpenairConverter(
        airspaces, borders, convert_DA_to_DB=False)
    openairs = converter.convert_airspaces()

    fp = open(args.output, 'wb')
    writer = aerofiles.openair.writer.Writer(fp)
    for openair in openairs:
        writer.write_record(openair)
        writer.write_line("")

    if args.borders:
        borders = converter.convert_borders()
        for border in borders:
            border["type"] = "airspace"
            border["floor"] = "GND"
            border["ceiling"] = "GND"
            writer.write_record(border)
            writer.write_line("")

    print(f'Converted {len(airspaces)} AIXM airspaces into {
          len(openairs)} openair airspaces')


if __name__ == '__main__':
    main()



import uuid
from dataclasses import dataclass, field
from datetime import time
from typing import List, Optional


from . import gml

# Airspace 1 ... n GeometryComponent
# GeometryComponent 1 ... 1 Volume
# Volume 1 ....  n Curves
#  -xor-
# Volume 1 ....  n Airspace (Dependency)


@dataclass
class ValDistanceVertical:
    """Vertical Distance Value"""
    value: float
    uom: str = "M"  # Unit of Measure (M=Meters, FT=Feet, FL=Flight Level)

    def __post_init__(self):
        if self.uom not in ["M", "FT", "FL"]:
            raise ValueError("UOM must be M, FT, or FL")


@dataclass
class AirspaceVolume:
    """AIXM AirspaceVolume Feature"""
    # Mandatory attributes
    gml_id: str = field(default_factory=lambda: f"asv_{uuid.uuid4().hex[:8]}")

    # Elements
    upper_limit: Optional[ValDistanceVertical] = None
    upper_limit_reference: Optional[str] = None           # AMSL, AGL, ...
    lower_limit: Optional[ValDistanceVertical] = None
    lower_limit_reference: Optional[str] = None           # AMSL, AGL, ...

    # Dependency relationships
    dependencies: List['AirspaceVolumeDependency'] = field(
        default_factory=list)
    curves: List['gml.Curve'] = field(default_factory=list)

    def add_curve(self, curve):
        self.curves.append(curve)

    def add_curves(self, curves):
        self.curves.extend(curves)

    def add_dependency(self, dependency):
        self.dependencies.append(dependency)


@dataclass
class AirspaceGeometryComponent:
    """AIXM AirspaceGeometryComponent - Association Class

    The role of the component in the airspace geometry. If the geometry of an
    airspace is composed of single volume, then the attributes of this
    association class may be left empty.
    """
    # Association attributes from AIXM spec
    operation: Optional[str] = None              # UNION, ADD, ...
    operation_sequence: Optional[int] = None

    # Reference to the actual geometry (Surface feature)
    volume: Optional['AirspaceVolume'] = None

    def set_operation(self, operation: str, sequence: int = 1):
        """Set the aggregation operation and sequence"""
        self.operation = operation
        self.operation_sequence = sequence

    def set_volume(self, volume: 'AirspaceVolume'):
        """Set the surface geometry for this component"""
        self.volume = volume


@dataclass
class AixmTimesheet:
    """
    AIXM Timesheet dataclass representing operational time periods.

    Attributes:
        time_reference: Time reference (UTC or LOCAL)
        start_date: Start date in MM-DD format
        end_date: End date in MM-DD format
        day: Day of the week
        day_til: End day of the week (optional)
        start_time: Start time in HH:MM format
        start_event: Start event (optional)
        start_time_relative_event: Start time relative to event (optional)
        start_event_interpretation: Start event interpretation (optional)
        end_time: End time in HH:MM format
        end_event: End event (optional)
        end_time_relative_event: End time relative to event (optional)
        end_event_interpretation: End event interpretation (optional)
        daylight_saving_adjust: Whether daylight saving time adjustments apply
        excluded: Whether this timesheet is excluded (optional)
    """
    time_reference: str  # "UTC" or "LOCAL"
    start_date: str      # Format: MM-DD
    end_date: str        # Format: MM-DD
    day: str             # "MON", "TUE", "WED", "THU", "FRI", "SAT", "SUN"
    start_time: str      # Format: HH:MM
    end_time: str        # Format: HH:MM
    daylight_saving_adjust: bool  # True for "YES", False for "NO"

    # Optional fields
    # day_til: Optional[str] = None
    # start_event: Optional[str] = None
    # start_time_relative_event: Optional[str] = None
    # start_event_interpretation: Optional[str] = None
    # end_event: Optional[str] = None
    # end_time_relative_event: Optional[str] = None
    # end_event_interpretation: Optional[str] = None
    # excluded: Optional[bool] = None

    @property
    def start_time_obj(self) -> time:
        """Get start_time as time object."""
        return self.start_time if isinstance(self.start_time, time) else self._parse_time(self.start_time)

    @property
    def end_time_obj(self) -> time:
        """Get end_time as time object."""
        return self.end_time if isinstance(self.end_time, time) else self._parse_time(self.end_time)

    def _parse_time(self, time_str: str) -> time:
        """Parse time string to time object."""
        hours, minutes = map(int, time_str.split(':'))
        return time(hours, minutes)


@dataclass
class AirspaceActivation:
    """
    AIXM Airspace Activation
    """
    activity: Optional[str] = None           # UAV, ...
    status: Optional[str] = None           # ACTIVE

    timesheets: List[AixmTimesheet] = field(default_factory=list)

    def add_timesheet(self, timesheet: AixmTimesheet):
        self.timesheets.append(timesheet)


@dataclass
class Airspace:
    """AIXM Airspace Feature"""
    # Mandatory attributes
    gml_id: str = field(default_factory=lambda: f"as_{uuid.uuid4().hex[:8]}")

    # Is this Airspace referenced by another AirspaceVolume?
    is_referenced: Optional[bool] = False

    # Elements
    designator: Optional[str] = None
    name: Optional[str] = None
    type_airspace: Optional[str] = None       # CTR, FIR, ...
    localtype_airspace: Optional[str] = None  # CTR, FIR, ...
    # designator_type: Optional[CodeAirspaceDesignator] = None

    # Controlled airspace properties
    # controlling_unit: Optional[str] = None  # Reference to Unit
    class_airspace: Optional[str] = None  # A, B, C, D, E, F, G

    # Activation properties
    activation: Optional[AirspaceActivation] = None

    # Airspace volumes
    components: List[AirspaceGeometryComponent] = field(default_factory=list)

    def add_component(self, component: AirspaceGeometryComponent):
        """Add an AirspaceVolume to this Airspace"""
        self.components.append(component)


@dataclass
class AirspaceVolumeDependency:
    """AIXM AirspaceVolumeDependency Feature"""
    # Mandatory attributes
    gml_id: str = field(default_factory=lambda: f"avd_{uuid.uuid4().hex[:8]}")

    # Dependency elements
    # FULL_GEOMETRY, OVERLAPS, ...
    type_dependency: Optional[str] = None
    # Reference to dependent airspace
    dependency_airspace: Optional['Airspace'] = None
    # Reference to dependent volume
    dependency_volume: Optional['AirspaceVolume'] = None
    xlink: Optional['str'] = None


@dataclass
class GeoBorder:
    """AIXM GeoBorder Feature"""
    # Mandatory attributes
    gml_id: str = field(default_factory=lambda: f"avd_{uuid.uuid4().hex[:8]}")
    name: str = None
    curve: Optional[gml.Curve] = None
